//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#ifndef _CVODEMANAGER_HXX_
#define _CVODEMANAGER_HXX_

#include "OdeManager.hxx"

#include <cvode/cvode_impl.h>
#include <cvode/cvode.h>            /* prototypes for CVODES fcts. and consts. */
#include <cvode/cvode_dense.h>     /* prototype for CVDense */
#include <cvode/cvode_direct.h>    /* prototypes for various DlsMat operations */

class CVODEManager final : public OdeManager
{
    public :

    CVODEManager(const std::wstring& callerName, const std::wstring& solverName, CVODEManager* prevManager = NULL) : OdeManager(callerName, solverName)
    {
        m_prevManager = prevManager;
        m_odeIsExtension = (prevManager != NULL);
    }

    ~CVODEManager() {
        if (m_cvode_mem != NULL) CVodeFree(&m_cvode_mem);
        m_cvode_mem = NULL;
    };
    
    void *getCVODE_mem()
    {
        return m_cvode_mem;
    }
    
    OdeManager *getPreviousManager()
    {
        return m_prevManager;
    }
    
    void *create();
    bool initialize(char *);
    bool computeIC(char *);
    void errHandler(int error_code, const char *module, const char *function, char *msg);
    bool setUserData();
    bool setInitStep();
    bool setMinStep();
    bool setMaxStep();    
    bool setStopTime();
    bool setTolerances();
    bool setMaxOrder();
    bool setEventFunction();
    bool setSolverAndJacobian(char *errorMsg);
    bool setHandlerFunction();
    bool getRootInfo(int *piRootsFound);
    bool getCurrentTime(double *pdblCurrTime);
    bool getCurrentStep(double *pdblCurrStep);
    bool getLastOrder(int *piLastOrder);
    bool getDky(double dblRefTime, int iOrder, N_Vector Y);
    solverReturnCode doStep(double dblFinalTime, N_Vector Y, N_Vector Yp, double *pdblTime, solverTaskCode iKind);
    void saveContinuousExtensionVectors();
    void getInterpVectors(double *pdblNS, int iOrderPlusOne, double dblt0, double dblTUser, double dblStep, double *pdblVect, double *pdblVectd);

    types::Struct *getStats();

    std::vector<std::wstring> getAvailableMethods()
    {
        std::vector<std::wstring> availableMethods = {L"adams",L"bdf"};
        return availableMethods;
    }

    int getMaxMethodOrder(std::wstring wstrMethod)
    {
        return wstrMethod == L"adams" ? 12 : 5;
    }    

    private :
    
    CVODEManager* m_prevManager;
    
    void *m_cvode_mem = NULL;
    
    long int m_incStat[9] = {0,0,0,0,0,0,0,0,0};
};

#endif
